//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Coord/CoordSystem2D.h
//! @brief     Defines interface CoordSystem2D and its subclasses.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif

#ifndef USER_API
#ifndef BORNAGAIN_DEVICE_COORD_COORDSYSTEM2D_H
#define BORNAGAIN_DEVICE_COORD_COORDSYSTEM2D_H

#include "Device/Coord/ICoordSystem.h"

class Beam;
class Direction;
class RectangularPixel;

//! Interface for objects that provide axis translations to different units for IDetector objects

class CoordSystem2D : public ICoordSystem {
public:
    CoordSystem2D(const Direction& direction, double wavelength = 0);
    CoordSystem2D(const std::vector<const IAxis*>& axes, const Direction& direction,
                  double wavelength = 0);
    ~CoordSystem2D() override = default;

    size_t rank() const override { return m_axes.size(); }

    double calculateMin(size_t i_axis, Coords units) const override;
    double calculateMax(size_t i_axis, Coords units) const override;
    size_t axisSize(size_t i_axis) const override;

    //! Returns list of units that are available for all 2D detectors.
    //! Further units may be added by child classes.
    std::vector<Coords> availableUnits() const override;

    IAxis* createConvertedAxis(size_t i_axis, Coords units) const override;

protected:
    CoordSystem2D(const CoordSystem2D& other);

    void addAxisData(std::string name, double min, double max, size_t nbins);

    std::vector<const IAxis*> m_axes;
    double m_wavelength;
    double m_alpha_i;
    double m_phi_i;

private:
    virtual double calculateValue(size_t i_axis, Coords units, double value) const = 0;
};


//! ICoordSystem class that handles the unit translations for spherical detectors
//! Its default units are radians for both axes

class SphericalCoords : public CoordSystem2D {
public:
    SphericalCoords(const std::vector<const IAxis*>& axes, const Direction& direction,
                    double wavelength);
    ~SphericalCoords() override;

    SphericalCoords* clone() const override;

    std::vector<Coords> availableUnits() const override;

    Coords defaultUnits() const override { return Coords::DEGREES; }

private:
    SphericalCoords(const SphericalCoords& other); //!< used by clone()
    double calculateValue(size_t i_axis, Coords units, double value) const override;
    std::vector<std::map<Coords, std::string>> createNameMaps() const override;
};


//! ICoordSystem class that handles the unit translations for rectangular detectors
//! Its default units are mm for both axes

class ImageCoords : public CoordSystem2D {
public:
    ImageCoords(const std::vector<const IAxis*>& axes,
                const RectangularPixel* regionOfInterestPixel, const Direction& direction,
                double wavelength);
    ~ImageCoords() override;

    ImageCoords* clone() const override;

    std::vector<Coords> availableUnits() const override;

    Coords defaultUnits() const override { return Coords::MM; }

private:
    ImageCoords(const ImageCoords& other); //!< used by clone()
    double calculateValue(size_t i_axis, Coords units, double value) const override;
    std::vector<std::map<Coords, std::string>> createNameMaps() const override;

    std::unique_ptr<const RectangularPixel> m_detector_pixel;
};


//! ICoordSystem class that handles the unit translations for off-specular simulations
//! with a spherical detector
//! Its default units are radians for both axes

class OffspecCoordinates : public CoordSystem2D {
public:
    OffspecCoordinates(const std::vector<const IAxis*>& axes, const Direction& direction);

    OffspecCoordinates* clone() const override;

    Coords defaultUnits() const override { return Coords::DEGREES; }

private:
    OffspecCoordinates(const OffspecCoordinates& other); //!< used by clone()
    double calculateValue(size_t i_axis, Coords units, double value) const override;
    std::vector<std::map<Coords, std::string>> createNameMaps() const override;
};


//! DepthProbeCoordinates class handles the unit translations for depth probe simulations
//! Its default units are radians for x-axis and nm for y-axis

class DepthProbeCoordinates : public CoordSystem2D {
public:
    DepthProbeCoordinates(const std::vector<const IAxis*>& axes, const Direction& direction,
                          double wavelength);
    ~DepthProbeCoordinates() override;

    DepthProbeCoordinates* clone() const override;

    std::vector<Coords> availableUnits() const override;

    Coords defaultUnits() const override { return Coords::DEGREES; }

private:
    DepthProbeCoordinates(const DepthProbeCoordinates& other); //!< used by clone()
    double calculateValue(size_t, Coords units, double value) const override;
    std::vector<std::map<Coords, std::string>> createNameMaps() const override;
};

#endif // BORNAGAIN_DEVICE_COORD_COORDSYSTEM2D_H
#endif // USER_API
