//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/AxisProperty.h
//! @brief     Defines class AxisProperty
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2022
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DESCRIPTOR_AXISPROPERTY_H
#define BORNAGAIN_GUI_MODEL_DESCRIPTOR_AXISPROPERTY_H

#include "Base/Axis/FixedBinAxis.h"
#include "GUI/Model/Descriptor/DoubleProperty.h"
#include <memory>

class Streamer;

//! Holds values which can be used to describe a FixedBinAxis.
//!
//! Use this as a member in your class to
//! * reduce dependency to axis items and axis classes
//! * simplify serialization
//! * easily have a UI supporting solution.
//!
//! This class is designed to work seamlessly with the serialization methods; just call it's
//! serialize for both reading and writing.
//! For UI integration just use AxisPropertyEditor.
//!
//! Do not forget to call all two init functions from within the containing class's constructor!

class AxisProperty {
public:
    explicit AxisProperty();

    DOUBLE_PROPERTY(min, Min);
    DOUBLE_PROPERTY(max, Max);

    void initMin(const QString& label, const QString& tooltip, double value,
                 const std::variant<QString, Unit>& unit,
                 const RealLimits& limits = RealLimits::nonnegative(), uint decimals = 3);
    void initMax(const QString& label, const QString& tooltip, double value,
                 const std::variant<QString, Unit>& unit,
                 const RealLimits& limits = RealLimits::nonnegative(), uint decimals = 3);

    //! Creates an axis with empty title
    std::unique_ptr<FixedBinAxis> createAxis(double scaleFactor) const;

    void rwAxisProperty(Streamer& s, const QString& tag);

    uint nbins() const { return m_nbins; }
    void setNbins(uint v) { m_nbins = v; }

private:
    uint m_nbins = 100;
};

// clang-format off

//! Add a member and a getter for an axis property
#define AXIS_PROPERTY(nameLower)                                                           \
protected:                                                                                 \
    AxisProperty m_##nameLower;                                                            \
public:                                                                                    \
    AxisProperty& nameLower() const { return *const_cast<AxisProperty*>(&m_##nameLower); }

// clang-format on

#endif // BORNAGAIN_GUI_MODEL_DESCRIPTOR_AXISPROPERTY_H
