//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/VectorProperty.cpp
//! @brief     Implements class VectorProperty
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Descriptor/VectorProperty.h"
#include "GUI/Support/XML/Serialize.h"
#include "GUI/Support/XML/UtilXML.h"
#include <QUuid>

namespace {
namespace Tag {

const QString X("X");
const QString Y("Y");
const QString Z("Z");

} // namespace Tag
} // namespace

using std::variant;

void VectorProperty::init(const QString& label, const QString& tooltip,
                          const variant<QString, Unit>& unit, const QString& uidPrefix)
{
    init(label, tooltip, R3(0.0, 0.0, 0.0), unit, 3 /*decimals*/, 0.01 /*step*/,
         RealLimits::limitless(), uidPrefix);
}

void VectorProperty::init(const QString& label, const QString& tooltip, const R3& value,
                          const std::variant<QString, Unit>& unit, uint decimals, double step,
                          const RealLimits& limits, const QString& uidPrefix)
{
    m_label = label;

    m_x.init("x", tooltip, value.x(), unit, decimals, step, limits, uidPrefix);
    m_y.init("y", tooltip, value.y(), unit, decimals, step, limits, uidPrefix);
    m_z.init("z", tooltip, value.z(), unit, decimals, step, limits, uidPrefix);

    if (uidPrefix.size() > 0)
        m_uid = uidPrefix + "/" + QUuid::createUuid().toString();
    else
        m_uid = QUuid::createUuid().toString();

    m_x.setUid(m_uid + "/x");
    m_y.setUid(m_uid + "/y");
    m_z.setUid(m_uid + "/z");
}

bool VectorProperty::operator==(const VectorProperty& other) const
{
    return (r3() == other.r3()) && (m_uid == other.m_uid) && (m_label == other.m_label);
}

void VectorProperty::rwProperty(Streamer& s, const QString& tag)
{
    if (QXmlStreamWriter* w = s.xmlWriter()) {
        w->writeStartElement(tag);
        Serialize::rwProperty(s, Tag::X, m_x);
        Serialize::rwProperty(s, Tag::Y, m_y);
        Serialize::rwProperty(s, Tag::Z, m_z);
        w->writeEndElement();
    } else if (QXmlStreamReader* r = s.xmlReader()) {
        r->readNextStartElement();
        s.assertCurrentTag(tag);
        Serialize::rwProperty(s, Tag::X, m_x);
        Serialize::rwProperty(s, Tag::Y, m_y);
        Serialize::rwProperty(s, Tag::Z, m_z);
        s.gotoEndElementOfTag(tag);
    } else
        ASSERT(0);
}
