//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/BeamDistributionItem.cpp
//! @brief     Implements class BeamDistributionItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Device/BeamDistributionItem.h"
#include "GUI/Model/CatDevice/DistributionItemCatalog.h"
#include "GUI/Support/XML/Serialize.h"
#include "Param/Distrib/Distributions.h"

std::unique_ptr<ParameterDistribution> BeamDistributionItem::getParameterDistributionForName(
    ParameterDistribution::WhichParameter which) const
{
    if (const auto* d = distribution()) {
        auto distribution1D = createDistribution1D();

        if (distribution1D) {
            const auto nbrOfSamples = d->numberOfSamples();
            const double sigmaFactor = d->hasSigmaFactor() ? d->sigmaFactor() : 0.0;
            const auto limits = d->limits().scaledLimits(scaleFactor());

            return std::make_unique<ParameterDistribution>(which, *distribution1D, nbrOfSamples,
                                                           sigmaFactor, limits);
        }
    }
    return {};
}

double BeamDistributionItem::meanValue() const
{
    std::unique_ptr<IDistribution1D> domainDistr = createDistribution1D();
    if (domainDistr)
        return domainDistr->mean() / scaleFactor();
    return dynamic_cast<SymmetricResolutionItem*>(distribution())->mean();
}

void BeamDistributionItem::resetToValue(double value)
{
    auto* d = new DistributionNoneItem();
    d->setMean(value);
    m_distribution.set(d);
}

double BeamDistributionItem::scaleFactor() const
{
    return 1.0;
}

DistributionItem* BeamDistributionItem::distribution() const
{
    return m_distribution.get();
}

SelectionDescriptor<DistributionItem*> BeamDistributionItem::distributionSelection() const
{
    return m_distribution;
}

void BeamDistributionItem::setDistribution(DistributionItem* d)
{
    m_distribution.set(d);
}

std::unique_ptr<IDistribution1D> BeamDistributionItem::createDistribution1D() const
{
    return distribution()->createDistribution(scaleFactor());
}
