//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/BeamItems.h
//! @brief     Defines BeamItem hierarchy
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DEVICE_BEAMITEMS_H
#define BORNAGAIN_GUI_MODEL_DEVICE_BEAMITEMS_H

#include "GUI/Model/Descriptor/SelectionDescriptor.h"
#include "GUI/Model/Descriptor/SelectionProperty.h"
#include "GUI/Model/Device/BeamAngleItems.h"
#include "GUI/Model/Device/BeamWavelengthItem.h"
#include "GUI/Model/Device/SpecularBeamInclinationItem.h"
#include <functional>
#include <heinz/Vectors3D.h>

class BasicAxisItem;
class Beam;
class BeamAzimuthalAngleItem;
class BeamWavelengthItem;
class FootprintItem;
class IAxis;

class BeamItem {
public:
    virtual ~BeamItem() = default;

    virtual void serialize(Streamer& s);

    DOUBLE_PROPERTY(intensity, Intensity);

    double wavelength() const;
    void setWavelength(double value);
    BeamWavelengthItem* wavelengthItem() const;

    virtual double getInclinationAngle() const = 0;
    virtual void setInclinationAngle(double value);
    virtual BeamDistributionItem* inclinationAngleItem() const;

    double getAzimuthalAngle() const;
    void setAzimuthalAngle(double value);
    BeamAzimuthalAngleItem* azimuthalAngleItem() const;

    std::shared_ptr<Beam> createBeam() const;

protected:
    BeamItem();

    template <typename T>
    void initWavelength();
    template <typename T>
    void initInclinationAngle();

    std::unique_ptr<BeamWavelengthItem> m_wavelengthItem;
    std::unique_ptr<BeamAzimuthalAngleItem> m_azimuthalAngleItem;
    std::unique_ptr<BeamDistributionItem> m_inclinationAngleItem;
};

class SpecularBeamItem : public BeamItem {
public:
    explicit SpecularBeamItem(const InstrumentItem* owningInstrument);
    void serialize(Streamer& s) override;

    double getInclinationAngle() const override;
    void setInclinationAngle(double value) override;
    SpecularBeamInclinationItem* inclinationAngleItem() const override;
    BasicAxisItem* inclinationAxis() const;

    FootprintItem* footprint() const;
    SelectionDescriptor<FootprintItem*> footprintSelection() const;
    void setGaussianFootprint(double value);
    void setSquareFootprint(double value);
    template <typename T>
    T* setFootprintType();

    void updateToData(const IAxis& axis, QString units);
    void updateWavelength();

private:
    SelectionProperty<FootprintItem*> m_footprint;
};

class GISASBeamItem : public BeamItem {
public:
    GISASBeamItem();

    double getInclinationAngle() const override;

    //! Explicitly set a getter function for the inclination angle.
    //!
    //! If a getter is set, it overrides the behavior in getInclinationAngle().
    //! Helpful if the angle cannot be determined from the internal beam parameters, but from
    //! external information.
    void setInclinationAngleGetter(std::function<double()> getter);

    std::function<double()> m_inclinationAngleGetter;
};

#endif // BORNAGAIN_GUI_MODEL_DEVICE_BEAMITEMS_H
