//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/FitParameterItem.cpp
//! @brief     Implements FitParameterItems family of classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Job/FitParameterItem.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Job/FitParameterLinkItem.h"
#include "GUI/Support/XML/UtilXML.h"

namespace {

namespace Tag {

const QString Value("Value");
const QString IsEnabled("IsEnabled");

const QString DisplayName("DisplayName");
const QString Type("Type");
const QString TypeItem("TypeItem");
const QString StartValue("StartValue");
const QString MinValue("MinValue");
const QString MaxValue("MaxValue");
const QString FitLinkItem("FitLinkItem");

} // namespace Tag

namespace Attrib {

const QString value("value");
const QString name("name");

} // namespace Attrib

} // namespace

FitTypeItem::FitTypeItem(const ComboProperty& type, QObject* parent)
    : QObject(parent)
    , m_type(type)
{
    setObjectName("Type");
}

ComboProperty FitTypeItem::type() const
{
    return m_type;
}

void FitTypeItem::setType(const ComboProperty& type)
{
    m_type = type;
}

void FitTypeItem::writeTo(QXmlStreamWriter* writer) const
{
    // type
    writer->writeStartElement(Tag::Type);
    m_type.writeTo(writer);
    writer->writeEndElement();
}

void FitTypeItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // type
        if (reader->name() == Tag::Type) {
            m_type.readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Type);

        } else
            reader->skipCurrentElement();
    }
}

//------------------------------------------------------------------------------------------------

FitDoubleItem::FitDoubleItem(double value, QObject* parent)
    : QObject(parent)
    , m_value(value)
    , m_decimals(3)
    , m_limits(RealLimits::limitless())
{
}

double FitDoubleItem::value() const
{
    return m_value;
}

void FitDoubleItem::setValue(double value)
{
    m_value = value;
}

RealLimits FitDoubleItem::limits() const
{
    return m_limits;
}

void FitDoubleItem::setLimits(const RealLimits& limits)
{
    m_limits = limits;
}

int FitDoubleItem::decimals() const
{
    return m_decimals;
}

void FitDoubleItem::setDecimals(int deciamls)
{
    m_decimals = deciamls;
}

void FitDoubleItem::writeTo(QXmlStreamWriter* writer) const
{
    // no need to write m_limits and m_decimals

    // value
    writer->writeStartElement(Tag::Value);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_value);
    writer->writeEndElement();
}

void FitDoubleItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement())
        readBasicTags(reader);
}

void FitDoubleItem::readBasicTags(QXmlStreamReader* reader)
{
    // no need to read m_limits and m_decimals

    // value
    if (reader->name() == Tag::Value) {
        GUI::Session::XML::readAttribute(reader, Attrib::value, &m_value);
        GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Value);

    } else
        reader->skipCurrentElement();
}

//------------------------------------------------------------------------------------------------

FitEditableDoubleItem::FitEditableDoubleItem(double value, bool isEnabled, QObject* parent)
    : FitDoubleItem(value, parent)
    , m_isEnabled(isEnabled)
{
}

bool FitEditableDoubleItem::isEnabled() const
{
    return m_isEnabled;
}

void FitEditableDoubleItem::setEnabled(bool enabled)
{
    m_isEnabled = enabled;
}

void FitEditableDoubleItem::writeTo(QXmlStreamWriter* writer) const
{
    FitDoubleItem::writeTo(writer);

    // is enabled
    writer->writeStartElement(Tag::IsEnabled);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_isEnabled);
    writer->writeEndElement();
}

void FitEditableDoubleItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // is enabled
        if (reader->name() == Tag::IsEnabled) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_isEnabled);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::IsEnabled);

            // types from base class
        } else {
            FitDoubleItem::readBasicTags(reader);
        }
    }
}

//------------------------------------------------------------------------------------------------

namespace {

ComboProperty fitParameterTypeCombo()
{
    QStringList tooltips = QStringList() << "Fixed at given value"
                                         << "Limited in the range [min, max]"
                                         << "Limited at lower bound [min, inf]"
                                         << "Limited at upper bound [-inf, max]"
                                         << "No limits imposed to parameter value";

    ComboProperty result;
    result << "fixed"
           << "limited"
           << "lower limited"
           << "upper limited"
           << "free";
    result.setCurrentValue("limited");
    result.setToolTips(tooltips);
    return result;
}

const double range_factor = 0.5;

} // namespace


FitParameterItem::FitParameterItem(QObject* parent)
    : QObject(parent)
    , m_typeItem(std::make_unique<FitTypeItem>(fitParameterTypeCombo(), this))
    , m_startValueItem(std::make_unique<FitDoubleItem>(0.0, this))
    , m_minItem(std::make_unique<FitEditableDoubleItem>(0.0, true, this))
    , m_maxItem(std::make_unique<FitEditableDoubleItem>(0.0, false, this))
{
    setObjectName("FitParameter");
    m_startValueItem->setObjectName("Value");
    m_minItem->setObjectName("Min");
    m_maxItem->setObjectName("Max");

    onTypeChange();
}

//! Inits P_MIN and P_MAX taking into account current value and external limits

void FitParameterItem::initMinMaxValues(const RealLimits& limits)
{
    double value = startValue();

    double dr(0);
    if (value == 0.0)
        dr = 1.0 * range_factor;
    else
        dr = std::abs(value) * range_factor;

    double min = value - dr;
    double max = value + dr;

    if (limits.hasLowerLimit() && min < limits.lowerLimit())
        min = limits.lowerLimit();

    if (limits.hasUpperLimit() && max > limits.upperLimit())
        max = limits.upperLimit();

    setMinimum(min);
    m_minItem->setLimits(limits);
    setMaximum(max);
    m_maxItem->setLimits(limits);

    m_startValueItem->setLimits(limits);
}

//! Constructs Limits corresponding to current GUI settings.

AttLimits FitParameterItem::attLimits() const
{
    if (isFixed())
        return AttLimits::fixed();
    if (isLimited())
        return AttLimits::limited(minimum(), maximum());
    if (isLowerLimited())
        return AttLimits::lowerLimited(minimum());
    if (isUpperLimited())
        return AttLimits::upperLimited(maximum());
    if (isFree())
        return AttLimits::limitless();
    ASSERT(0);
}

bool FitParameterItem::isValid() const
{
    if (isFixed() || isFree())
        return true;

    if (isLowerLimited())
        return minimum() <= startValue();
    if (isUpperLimited())
        return startValue() <= maximum();
    return minimum() <= startValue() && startValue() <= maximum();
}

QString FitParameterItem::displayName() const
{
    return m_displayName;
}

void FitParameterItem::setDisplayName(QString displayName)
{
    m_displayName = displayName;
}

double FitParameterItem::startValue() const
{
    return m_startValueItem->value();
}

void FitParameterItem::setStartValue(double start_value)
{
    m_startValueItem->setValue(start_value);
}

QObject* FitParameterItem::startValueItem() const
{
    return m_startValueItem.get();
}

double FitParameterItem::minimum() const
{
    return m_minItem->value();
}

void FitParameterItem::setMinimum(double minimum)
{
    m_minItem->setValue(minimum);
}

QObject* FitParameterItem::minimumItem() const
{
    return m_minItem.get();
}

double FitParameterItem::maximum() const
{
    return m_maxItem->value();
}

void FitParameterItem::setMaximum(double maximum)
{
    m_maxItem->setValue(maximum);
}

QObject* FitParameterItem::maximumItem() const
{
    return m_maxItem.get();
}

FitParameterLinkItem* FitParameterItem::addLink(const QString& title, const QString& link)
{
    FitParameterLinkItem* newLink = new FitParameterLinkItem(this);
    m_links.emplace_back(newLink);

    newLink->setTitle(title);
    newLink->setLink(link);
    return newLink;
}

void FitParameterItem::removeLink(const QString& link)
{
    for (FitParameterLinkItem* linkItem : m_links)
        if (linkItem->link() == link)
            m_links.delete_element(linkItem);
}

QVector<FitParameterLinkItem*> FitParameterItem::linkItems() const
{
    return QVector<FitParameterLinkItem*>(m_links.begin(), m_links.end());
}

QStringList FitParameterItem::links() const
{
    QStringList links;
    for (FitParameterLinkItem* linkItem : linkItems())
        links << linkItem->link();
    return links;
}

void FitParameterItem::setTypeCombo(const ComboProperty& type)
{
    m_typeItem->setType(type);
    onTypeChange();
}

QString FitParameterItem::currentType() const
{
    return m_typeItem->type().currentValue();
}

QObject* FitParameterItem::typeItem() const
{
    return m_typeItem.get();
}

void FitParameterItem::writeTo(QXmlStreamWriter* writer) const
{
    // displayName
    writer->writeStartElement(Tag::DisplayName);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_displayName);
    writer->writeEndElement();

    // type
    writer->writeStartElement(Tag::TypeItem);
    m_typeItem->writeTo(writer);
    writer->writeEndElement();

    // start value
    writer->writeStartElement(Tag::StartValue);
    m_startValueItem->writeTo(writer);
    writer->writeEndElement();

    // min value
    writer->writeStartElement(Tag::MinValue);
    m_minItem->writeTo(writer);
    writer->writeEndElement();

    // max value
    writer->writeStartElement(Tag::MaxValue);
    m_maxItem->writeTo(writer);
    writer->writeEndElement();

    // parameter links
    for (const auto* fitLink : linkItems()) {
        writer->writeStartElement(Tag::FitLinkItem);
        writer->writeAttribute(Attrib::name, fitLink->title());
        fitLink->writeTo(writer);
        writer->writeEndElement();
    }
}

void FitParameterItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // displayName
        if (reader->name() == Tag::DisplayName) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_displayName);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::DisplayName);

            // type
        } else if (reader->name() == Tag::TypeItem) {
            m_typeItem->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::TypeItem);

            // start value
        } else if (reader->name() == Tag::StartValue) {
            m_startValueItem->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::StartValue);

            // min value
        } else if (reader->name() == Tag::MinValue) {
            m_minItem->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::MinValue);

            // max value
        } else if (reader->name() == Tag::MaxValue) {
            m_maxItem->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::MaxValue);

            // parameter links
        } else if (reader->name() == Tag::FitLinkItem) {
            addLink("", "")->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::FitLinkItem);

        } else
            reader->skipCurrentElement();
    }
}

//! Enables/disables min, max properties on FitParameterItem's type

void FitParameterItem::onTypeChange()
{
    if (isFixed()) {
        setLimitEnabled(m_minItem.get(), false);
        setLimitEnabled(m_maxItem.get(), false);
    }

    else if (isLimited()) {
        setLimitEnabled(m_minItem.get(), true);
        setLimitEnabled(m_maxItem.get(), true);
    }

    else if (isLowerLimited()) {
        setLimitEnabled(m_minItem.get(), true);
        setLimitEnabled(m_maxItem.get(), false);
    }

    else if (isUpperLimited()) {
        setLimitEnabled(m_minItem.get(), false);
        setLimitEnabled(m_maxItem.get(), true);
    }

    else if (isFree()) {
        setLimitEnabled(m_minItem.get(), false);
        setLimitEnabled(m_maxItem.get(), false);
    }
}

//! Set limit property with given name to the enabled state

void FitParameterItem::setLimitEnabled(FitEditableDoubleItem* propertyItem, bool enabled)
{
    ASSERT(propertyItem);
    propertyItem->setEnabled(enabled);
}

bool FitParameterItem::isLimited() const
{
    return currentType() == "limited";
}

bool FitParameterItem::isFree() const
{
    return currentType() == "free";
}

bool FitParameterItem::isLowerLimited() const
{
    return currentType() == "lower limited";
}

bool FitParameterItem::isUpperLimited() const
{
    return currentType() == "upper limited";
}

bool FitParameterItem::isFixed() const
{
    return currentType() == "fixed";
}
