//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/JobItem.h
//! @brief     Defines class JobItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_JOB_JOBITEM_H
#define BORNAGAIN_GUI_MODEL_JOB_JOBITEM_H

#include "GUI/Model/BaseItem/JobRealBase.h"
#include "GUI/Support/Data/JobStatus.h" // enum cannot be forward declared
#include "GUI/Support/Data/SimulationOptionsItem.h"

#include <QDateTime>
#include <optional>

class DataItem;
class FitParameterContainerItem;
class FitSuiteItem;
class Instrument2DItem;
class InstrumentItem;
class InstrumentItems;
class IntensityDataItem;
class MultiLayerItem;
class ParameterContainerItem;
class ParameterTreeItems;
class RealItem;
class SimulationResult;

class BA_CORE_API_ JobItem : public JobRealBase {
    Q_OBJECT

public:
    static constexpr auto M_TYPE{"JobItem"};

    JobItem();
    ~JobItem();

    // job properties

    QString identifier() const;
    void setIdentifier(const QString& identifier);

    QString jobName() const;
    void setJobName(const QString& name);

    JobStatus status() const;
    void setStatus(JobStatus status);

    bool isIdle() const;
    bool isRunning() const;
    bool isCompleted() const;
    bool isCanceled() const;
    bool isFailed() const;
    bool isFitting() const;
    bool isValidForFitting();

    QDateTime beginTime() const;
    void setBeginTime(const QDateTime& begin_time);

    QDateTime endTime() const;
    void setEndTime(const QDateTime& end_time);

    //! if begin and end time are both available the duration in ms, otherwise empty
    std::optional<size_t> duration() const;

    QString comments() const;
    void setComments(const QString& comments);

    int progress() const;
    void setProgress(int progress);

    bool runImmediately() const;
    bool runInBackground() const;

    QString presentationType() const;
    void setPresentationType(const QString& type);

    QString activity() const;
    void setActivity(const QString& activity);

    bool isSpecularJob() const;
    bool isIntensityJob() const;

    // sample

    MultiLayerItem* sampleItem();
    void copySampleIntoJob(const MultiLayerItem* sample);

    // instrument

    InstrumentItem* instrumentItem() const;
    Instrument2DItem* instrument2DItem();
    void copyInstrumentIntoJob(const InstrumentItem* instrument);

    // simulation

    const SimulationOptionsItem& simulationOptionsItem() const;
    void copySimulationOptionsIntoJob(const SimulationOptionsItem& options);
    void setResults(const SimulationResult& result);

    // fitting

    FitSuiteItem* fitSuiteItem();
    FitSuiteItem* createFitSuiteItem();
    void createFitContainers();

    ParameterContainerItem* parameterContainerItem();
    FitParameterContainerItem* fitParameterContainerItem();

    // data

    void createSumulatedDataItem();
    IntensityDataItem* intensityDataItem();
    DataItem* simulatedDataItem();

    void createDiffDataItem();
    DataItem* diffDataItem();

    RealItem* createRealItem();
    void copyRealItemIntoJob(const RealItem* srcRealItem);
    RealItem* realItem();
    void adjustReaDataToJobInstrument();

    // write/read

    void writeTo(QXmlStreamWriter* writer) const;
    void readFrom(QXmlStreamReader* reader);

signals:
    void jobNameChanged(const QString& name);
    void jobStatusChanged(JobStatus status);
    void jobBeginTimeChanged(const QDateTime& begin_time);
    void jobEndTimeChanged(const QDateTime& end_time);
    void jobCommentsChanged(const QString& comments);
    void jobProgressChanged(int progress);

private:
    void updateDataFileName();
    DataItem* createNewDataItem();
    void cropRealData();

private:
    std::unique_ptr<SimulationOptionsItem> m_simulationOptionsItem;
    std::unique_ptr<ParameterContainerItem> m_parameterContainer;
    std::unique_ptr<MultiLayerItem> m_sampleItem;
    std::unique_ptr<InstrumentItem>
        m_instrument; // TODO: Why own the pointer?
                      // ANSWER: job may exist even if the original instrument is deleted.
                      // For the uniformity purpose, it can be just "InstrumentItem m_instrument;"
    QString m_name;
    QString m_identifier;
    JobStatus m_status = JobStatus::Idle;
    QString m_comments;
    QString m_activity;
    QString m_presentationType;
    QDateTime m_beginTime;
    QDateTime m_endTime;
    int m_progress;
    std::unique_ptr<DataItem> m_simulatedDataItem; // TODO: Can be replaced by std::variant
    std::unique_ptr<DataItem> m_diffDataItem;      // TODO: Can be replaced by std::variant
    std::unique_ptr<RealItem> m_realItem;          // TODO: Can be replaced by optional
    std::unique_ptr<FitSuiteItem> m_fitSuiteItem;  // TODO: Can be replaced by optional
};

#endif // BORNAGAIN_GUI_MODEL_JOB_JOBITEM_H
