//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/FormFactorItems.cpp
//! @brief     Implements FormFactorItems classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/FormFactorItems.h"
#include "Base/Const/Units.h"
#include "Sample/HardParticle/HardParticles.h"

namespace {
namespace Tag {

const QString Length("Length");
const QString Width("Width");
const QString Height("Height");
const QString Alpha("Alpha");
const QString Radius("Radius");
const QString BaseEdge("BaseEdge");
const QString HeightRatio("HeightRatio");
const QString Edge("Edge");
const QString RadiusX("RadiusX");
const QString RadiusY("RadiusY");
const QString Asymmetry("Asymmetry");
const QString RemovedLength("RemovedLength");
const QString RemovedTop("RemovedTop");
const QString HeightFlattening("HeightFlattening");
const QString SliceBottom("SliceBottom");
const QString SliceTop("SliceTop");

} // namespace Tag
} // namespace

/* ------------------------------------------------ */

Pyramid2Item::Pyramid2Item()
{
    m_length.init("Length", "Length of the rectangular base", 16.0, Unit::nanometer, "length");
    m_width.init("Width", "Width of the rectangular base", 16.0, Unit::nanometer, "width");
    m_height.init("Height", "Height of pyramid", 16.0, Unit::nanometer, "height");
    m_alpha.init("Alpha", "Dihedral angle between base and facet", 80.0, Unit::degree, "alpha");
}

std::unique_ptr<IFormFactor> Pyramid2Item::createFormFactor() const
{
    return std::make_unique<Pyramid2>(m_length.value(), m_width.value(), m_height.value(),
                                      m_alpha.value() * Units::deg);
}

void Pyramid2Item::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::Width, m_width);
    Serialize::rwProperty(s, Tag::Height, m_height);
    Serialize::rwProperty(s, Tag::Alpha, m_alpha);
}

/* ------------------------------------------------ */

BarGaussItem::BarGaussItem()
{
    m_length.init("Length", "Length of the base", 16.0, Unit::nanometer, "length");
    m_width.init("Width", "Width of the base", 16.0, Unit::nanometer, "width");
    m_height.init("Height", "Height of the box", 16.0, Unit::nanometer, "height");
}

std::unique_ptr<IFormFactor> BarGaussItem::createFormFactor() const
{
    return std::make_unique<BarGauss>(m_length.value(), m_width.value(), m_height.value());
}

void BarGaussItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::Width, m_width);
    Serialize::rwProperty(s, Tag::Height, m_height);
}

/* ------------------------------------------------ */

BarLorentzItem::BarLorentzItem()
{
    m_length.init("Length", "Length of the base", 16.0, Unit::nanometer, "length");
    m_width.init("Width", "Width of the base", 16.0, Unit::nanometer, "width");
    m_height.init("Height", "Height of the box", 16.0, Unit::nanometer, "height");
}

std::unique_ptr<IFormFactor> BarLorentzItem::createFormFactor() const
{
    return std::make_unique<BarLorentz>(m_length.value(), m_width.value(), m_height.value());
}

void BarLorentzItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::Width, m_width);
    Serialize::rwProperty(s, Tag::Height, m_height);
}

/* ------------------------------------------------ */

BoxItem::BoxItem()
{
    m_length.init("Length", "Length of the base", 16.0, Unit::nanometer, "length");
    m_width.init("Width", "Width of the base", 16.0, Unit::nanometer, "width");
    m_height.init("Height", "Height of the box", 16.0, Unit::nanometer, "height");
}

std::unique_ptr<IFormFactor> BoxItem::createFormFactor() const
{
    return std::make_unique<Box>(m_length.value(), m_width.value(), m_height.value());
}

void BoxItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::Width, m_width);
    Serialize::rwProperty(s, Tag::Height, m_height);
}

/* ------------------------------------------------ */

ConeItem::ConeItem()
{
    m_radius.init("Radius", "Radius of the base", 8.0, Unit::nanometer, "radius");
    m_height.init("Height", "Height of the cone", 16.0, Unit::nanometer, "height");
    m_alpha.init("Alpha", "Angle between the base and the side surface", 65.0, Unit::degree,
                 "alpha");
}

std::unique_ptr<IFormFactor> ConeItem::createFormFactor() const
{
    return std::make_unique<Cone>(m_radius.value(), m_height.value(), m_alpha.value() * Units::deg);
}

void ConeItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Radius, m_radius);
    Serialize::rwProperty(s, Tag::Height, m_height);
    Serialize::rwProperty(s, Tag::Alpha, m_alpha);
}

/* ------------------------------------------------ */

Pyramid6Item::Pyramid6Item()
{
    m_baseEdge.init("Base edge", "Edge of the regular hexagonal base", 8.0, Unit::nanometer,
                    "baseEdge");
    m_height.init("Height", "Height of a truncated pyramid", 16.0, Unit::nanometer, "height");
    m_alpha.init("Alpha", "Dihedral angle between base and facet", 70.0, Unit::degree, "alpha");
}

std::unique_ptr<IFormFactor> Pyramid6Item::createFormFactor() const
{
    return std::make_unique<Pyramid6>(m_baseEdge.value(), m_height.value(),
                                      m_alpha.value() * Units::deg);
}

void Pyramid6Item::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::BaseEdge, m_baseEdge);
    Serialize::rwProperty(s, Tag::Height, m_height);
    Serialize::rwProperty(s, Tag::Alpha, m_alpha);
}

/* ------------------------------------------------ */

Bipyramid4Item::Bipyramid4Item()
{
    m_length.init("Length", "Side length of the common square base", 12.0, Unit::nanometer,
                  "length");

    m_height.init("Height", "Height of the lower pyramid", 16.0, Unit::nanometer, "height");

    m_heightRatio.init("Height ratio", "Ratio of heights of top to bottom pyramids", 0.7,
                       Unit::unitless, 3, RealLimits::lowerLimited(0.0), "heightRatio");

    m_alpha.init("Alpha", "Dihedral angle between base and facets", 70.0, Unit::degree, "alpha");
}

std::unique_ptr<IFormFactor> Bipyramid4Item::createFormFactor() const
{
    return std::make_unique<Bipyramid4>(m_length.value(), m_height.value(), m_heightRatio.value(),
                                        m_alpha.value() * Units::deg);
}

void Bipyramid4Item::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::Height, m_height);
    Serialize::rwProperty(s, Tag::HeightRatio, m_heightRatio);
    Serialize::rwProperty(s, Tag::Alpha, m_alpha);
}

/* ------------------------------------------------ */

CylinderItem::CylinderItem()
{
    m_radius.init("Radius", "Radius of the circular base", 8.0, Unit::nanometer, "radius");
    m_height.init("Height", "Height of the cylinder", 16.0, Unit::nanometer, "height");
}

std::unique_ptr<IFormFactor> CylinderItem::createFormFactor() const
{
    return std::make_unique<Cylinder>(m_radius.value(), m_height.value());
}

void CylinderItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Radius, m_radius);
    Serialize::rwProperty(s, Tag::Height, m_height);
}

/* ------------------------------------------------ */

EllipsoidalCylinderItem::EllipsoidalCylinderItem()
{
    m_radiusX.init("Radius x", "Radius of the ellipse base in the x-direction", 5.0,
                   Unit::nanometer, "radiusX");
    m_radiusY.init("Radius y", "Radius of the ellipse base in the y-direction", 8.0,
                   Unit::nanometer, "radiusY");
    m_height.init("Height", "Height of the ellipsoidal cylinder", 16.0, Unit::nanometer, "height");
}

std::unique_ptr<IFormFactor> EllipsoidalCylinderItem::createFormFactor() const
{
    return std::make_unique<EllipsoidalCylinder>(m_radiusX.value(), m_radiusY.value(),
                                                 m_height.value());
}

void EllipsoidalCylinderItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::RadiusX, m_radiusX);
    Serialize::rwProperty(s, Tag::RadiusY, m_radiusY);
    Serialize::rwProperty(s, Tag::Height, m_height);
}

/* ------------------------------------------------ */

SphereItem::SphereItem()
{
    m_radius.init("Radius", "Radius of the sphere", 8.0, Unit::nanometer, "radius");
}

std::unique_ptr<IFormFactor> SphereItem::createFormFactor() const
{
    return std::make_unique<Sphere>(m_radius.value());
}

void SphereItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Radius, m_radius);
}

/* ------------------------------------------------ */

SpheroidItem::SpheroidItem()
{
    m_radius.init("Radius", "Radius of the circular cross section", 8.0, Unit::nanometer, "radius");
    m_height.init("Height", "Height of the full spheroid", 16.0, Unit::nanometer, "height");
}

std::unique_ptr<IFormFactor> SpheroidItem::createFormFactor() const
{
    return std::make_unique<Spheroid>(m_radius.value(), m_height.value());
}

void SpheroidItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Radius, m_radius);
    Serialize::rwProperty(s, Tag::Height, m_height);
}

/* ------------------------------------------------ */

HemiEllipsoidItem::HemiEllipsoidItem()
{
    m_radiusX.init("Radius x", "Radius of the ellipse base in the x-direction", 8.0,
                   Unit::nanometer, "radiusX");
    m_radiusY.init("Radius y", "Radius of the ellipse base in the y-direction", 6.0,
                   Unit::nanometer, "radiusY");
    m_height.init("Height", "Height of the hemi ellipsoid", 16.0, Unit::nanometer, "height");
}

std::unique_ptr<IFormFactor> HemiEllipsoidItem::createFormFactor() const
{
    return std::make_unique<HemiEllipsoid>(m_radiusX.value(), m_radiusY.value(), m_height.value());
}

void HemiEllipsoidItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::RadiusX, m_radiusX);
    Serialize::rwProperty(s, Tag::RadiusY, m_radiusY);
    Serialize::rwProperty(s, Tag::Height, m_height);
}

/* ------------------------------------------------ */

Prism3Item::Prism3Item()
{
    m_baseEdge.init("Base edge", "Length of the base edge", 14.0, Unit::nanometer, "baseEdge");
    m_height.init("Height", "Height", 16.0, Unit::nanometer, "height");
}

std::unique_ptr<IFormFactor> Prism3Item::createFormFactor() const
{
    return std::make_unique<Prism3>(m_baseEdge.value(), m_height.value());
}

void Prism3Item::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::BaseEdge, m_baseEdge);
    Serialize::rwProperty(s, Tag::Height, m_height);
}

/* ------------------------------------------------ */

Prism6Item::Prism6Item()
{
    m_baseEdge.init("Base edge", "Length of the hexagonal base", 8.0, Unit::nanometer, "baseEdge");
    m_height.init("Height", "Height", 16.0, Unit::nanometer, "height");
}

std::unique_ptr<IFormFactor> Prism6Item::createFormFactor() const
{
    return std::make_unique<Prism6>(m_baseEdge.value(), m_height.value());
}

void Prism6Item::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::BaseEdge, m_baseEdge);
    Serialize::rwProperty(s, Tag::Height, m_height);
}

/* ------------------------------------------------ */

Pyramid4Item::Pyramid4Item()
{
    m_baseEdge.init("Base edge", "Length of the square base", 16.0, Unit::nanometer, "baseEdge");
    m_height.init("Height", "Height of the pyramid", 16.0, Unit::nanometer, "height");
    m_alpha.init("Alpha", "Dihedral angle between the base and a side face", 65.0, Unit::degree,
                 "alpha");
}

std::unique_ptr<IFormFactor> Pyramid4Item::createFormFactor() const
{
    return std::make_unique<Pyramid4>(m_baseEdge.value(), m_height.value(),
                                      m_alpha.value() * Units::deg);
}

void Pyramid4Item::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::BaseEdge, m_baseEdge);
    Serialize::rwProperty(s, Tag::Height, m_height);
    Serialize::rwProperty(s, Tag::Alpha, m_alpha);
}

/* ------------------------------------------------ */

CosineRippleBoxItem::CosineRippleBoxItem()
{
    m_length.init("Length", "Length of the rectangular base", 16.0, Unit::nanometer, "length");
    m_width.init("Width", "Width of the rectangular base", 16.0, Unit::nanometer, "width");
    m_height.init("Height", "Height of the ripple", 16.0, Unit::nanometer, "height");
}

std::unique_ptr<IFormFactor> CosineRippleBoxItem::createFormFactor() const
{
    return std::make_unique<CosineRippleBox>(m_length.value(), m_width.value(), m_height.value());
}

void CosineRippleBoxItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::Width, m_width);
    Serialize::rwProperty(s, Tag::Height, m_height);
}

/* ------------------------------------------------ */

CosineRippleGaussItem::CosineRippleGaussItem()
{
    m_length.init("Length", "Length of the rectangular base", 16.0, Unit::nanometer, "length");
    m_width.init("Width", "Width of the rectangular base", 16.0, Unit::nanometer, "width");
    m_height.init("Height", "Height of the ripple", 16.0, Unit::nanometer, "height");
}

std::unique_ptr<IFormFactor> CosineRippleGaussItem::createFormFactor() const
{
    return std::make_unique<CosineRippleGauss>(m_length.value(), m_width.value(), m_height.value());
}

void CosineRippleGaussItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::Width, m_width);
    Serialize::rwProperty(s, Tag::Height, m_height);
}

/* ------------------------------------------------ */

CosineRippleLorentzItem::CosineRippleLorentzItem()
{
    m_length.init("Length", "Length of the rectangular base", 16.0, Unit::nanometer, "length");
    m_width.init("Width", "Width of the rectangular base", 16.0, Unit::nanometer, "width");
    m_height.init("Height", "Height of the ripple", 16.0, Unit::nanometer, "height");
}

std::unique_ptr<IFormFactor> CosineRippleLorentzItem::createFormFactor() const
{
    return std::make_unique<CosineRippleLorentz>(m_length.value(), m_width.value(),
                                                 m_height.value());
}

void CosineRippleLorentzItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::Width, m_width);
    Serialize::rwProperty(s, Tag::Height, m_height);
}

/* ------------------------------------------------ */

SawtoothRippleBoxItem::SawtoothRippleBoxItem()
{
    m_length.init("Length", "Length of the rectangular base", 16.0, Unit::nanometer, "length");
    m_width.init("Width", "Width of the rectangular base", 16.0, Unit::nanometer, "width");
    m_height.init("Height", "Height of the ripple", 16.0, Unit::nanometer, "height");
    m_asymmetry.init("Asymmetry", "Asymmetry length of the triangular profile", 3.0,
                     Unit::nanometer, "asymmetry");
}

std::unique_ptr<IFormFactor> SawtoothRippleBoxItem::createFormFactor() const
{
    return std::make_unique<SawtoothRippleBox>(m_length.value(), m_width.value(), m_height.value(),
                                               m_asymmetry.value());
}

void SawtoothRippleBoxItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::Width, m_width);
    Serialize::rwProperty(s, Tag::Height, m_height);
    Serialize::rwProperty(s, Tag::Asymmetry, m_asymmetry);
}

/* ------------------------------------------------ */

SawtoothRippleGaussItem::SawtoothRippleGaussItem()
{
    m_length.init("Length", "Length of the rectangular base", 16.0, Unit::nanometer, "length");
    m_width.init("Width", "Width of the rectangular base", 16.0, Unit::nanometer, "width");
    m_height.init("Height", "Height of the ripple", 16.0, Unit::nanometer, "height");
    m_asymmetry.init("Asymmetry", "Asymmetry length of the triangular profile", 3.0,
                     Unit::nanometer, "asymmetry");
}

std::unique_ptr<IFormFactor> SawtoothRippleGaussItem::createFormFactor() const
{
    return std::make_unique<SawtoothRippleGauss>(m_length.value(), m_width.value(),
                                                 m_height.value(), m_asymmetry.value());
}

void SawtoothRippleGaussItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::Width, m_width);
    Serialize::rwProperty(s, Tag::Height, m_height);
    Serialize::rwProperty(s, Tag::Asymmetry, m_asymmetry);
}

/* ------------------------------------------------ */

SawtoothRippleLorentzItem::SawtoothRippleLorentzItem()
{
    m_length.init("Length", "Length of the rectangular base", 16.0, Unit::nanometer, "length");
    m_width.init("Width", "Width of the rectangular base", 16.0, Unit::nanometer, "width");
    m_height.init("Height", "Height of the ripple", 16.0, Unit::nanometer, "height");
    m_asymmetry.init("Asymmetry", "Asymmetry length of the triangular profile", 3.0,
                     Unit::nanometer, "asymmetry");
}

std::unique_ptr<IFormFactor> SawtoothRippleLorentzItem::createFormFactor() const
{
    return std::make_unique<SawtoothRippleLorentz>(m_length.value(), m_width.value(),
                                                   m_height.value(), m_asymmetry.value());
}

void SawtoothRippleLorentzItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::Width, m_width);
    Serialize::rwProperty(s, Tag::Height, m_height);
    Serialize::rwProperty(s, Tag::Asymmetry, m_asymmetry);
}

/* ------------------------------------------------ */

Pyramid3Item::Pyramid3Item()
{
    m_baseEdge.init("Base edge", "Length of one edge of the equilateral triangular base", 14.0,
                    Unit::nanometer, "baseEdge");
    m_height.init("Height", "Height of the tetrahedron", 16.0, Unit::nanometer, "height");
    m_alpha.init("Alpha", "Dihedral angle between base and facet", 80.0, Unit::degree, "alpha");
}

std::unique_ptr<IFormFactor> Pyramid3Item::createFormFactor() const
{
    return std::make_unique<Pyramid3>(m_baseEdge.value(), m_height.value(),
                                      m_alpha.value() * Units::deg);
}

void Pyramid3Item::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::BaseEdge, m_baseEdge);
    Serialize::rwProperty(s, Tag::Height, m_height);
    Serialize::rwProperty(s, Tag::Alpha, m_alpha);
}

/* ------------------------------------------------ */

TruncatedCubeItem::TruncatedCubeItem()
{
    m_length.init("Length", "Length of the full cube's edge", 16.0, Unit::nanometer, "length");
    m_removedLength.init("Removed length", "Removed length from each corner of the cube", 4.0,
                         Unit::nanometer, "removedLength");
}

std::unique_ptr<IFormFactor> TruncatedCubeItem::createFormFactor() const
{
    return std::make_unique<TruncatedCube>(m_length.value(), m_removedLength.value());
}

void TruncatedCubeItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::RemovedLength, m_removedLength);
}

/* ------------------------------------------------ */

TruncatedSphereItem::TruncatedSphereItem()
{
    m_radius.init("Radius", "Radius of the truncated sphere", 8.0, Unit::nanometer, "radius");
    m_height.init("Height", "Height of the truncated sphere", 16.0, Unit::nanometer, "height");
    m_removedTop.init("Delta height", "Height of the removed top cap", 16.0, Unit::nanometer,
                      "removedTop");
}

std::unique_ptr<IFormFactor> TruncatedSphereItem::createFormFactor() const
{
    return std::make_unique<TruncatedSphere>(m_radius.value(), m_height.value(),
                                             m_removedTop.value());
}

void TruncatedSphereItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Radius, m_radius);
    Serialize::rwProperty(s, Tag::Height, m_height);
    Serialize::rwProperty(s, Tag::RemovedTop, m_removedTop);
}

/* ------------------------------------------------ */

TruncatedSpheroidItem::TruncatedSpheroidItem()
{
    m_radius.init("Radius", "Radius of the truncated spheroid", 8.0, Unit::nanometer, "radius");
    m_height.init("Height", "Height of the truncated spheroid", 16.0, Unit::nanometer, "height");
    m_heightFlattening.init(
        "Height flattening",
        "Ratio of the height of the corresponding full spheroid to its diameter", 2.0,
        Unit::unitless, "heightFlattening");

    m_removedTop.init("Delta height", "Height of the removed top cap", 0.0, Unit::nanometer,
                      "removedTop");
}

std::unique_ptr<IFormFactor> TruncatedSpheroidItem::createFormFactor() const
{
    return std::make_unique<TruncatedSpheroid>(m_radius.value(), m_height.value(),
                                               m_heightFlattening.value(), m_removedTop.value());
}

void TruncatedSpheroidItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Radius, m_radius);
    Serialize::rwProperty(s, Tag::Height, m_height);
    Serialize::rwProperty(s, Tag::HeightFlattening, m_heightFlattening);
    Serialize::rwProperty(s, Tag::RemovedTop, m_removedTop);
}

/* ------------------------------------------------ */

CantellatedCubeItem::CantellatedCubeItem()
{
    m_length.init("Length", "Length of the full cube's edge", 16.0, Unit::nanometer, "length");
    m_removedLength.init("Removed length", "Removed length from each edge and vertice of the cube",
                         4.0, Unit::nanometer, "removedLength");
}

std::unique_ptr<IFormFactor> CantellatedCubeItem::createFormFactor() const
{
    return std::make_unique<CantellatedCube>(m_length.value(), m_removedLength.value());
}

void CantellatedCubeItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::RemovedLength, m_removedLength);
}

/* ------------------------------------------------ */

HorizontalCylinderItem::HorizontalCylinderItem()
{
    m_radius.init("Radius", "Radius of the horizontal cylinder", 8.0, Unit::nanometer, "radius");
    m_length.init("Length", "Length of the horizontal cylinder", 16.0, Unit::nanometer, "length");
    m_sliceBottom.init("Bottom boundary", "Position of the lower boundary relative to the center",
                       -4.1, Unit::nanometer, "sliceBottom");
    m_sliceTop.init("Top boundary", "Position of the upper boundary relative to the center", +5.2,
                    Unit::nanometer, "sliceTop");
}

std::unique_ptr<IFormFactor> HorizontalCylinderItem::createFormFactor() const
{
    return std::make_unique<HorizontalCylinder>(m_radius.value(), m_length.value(),
                                                m_sliceBottom.value(), m_sliceTop.value());
}

void HorizontalCylinderItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Radius, m_radius);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::SliceBottom, m_sliceBottom);
    Serialize::rwProperty(s, Tag::SliceTop, m_sliceTop);
}

/* ------------------------------------------------ */

void PlatonicItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Edge, m_edge);
}

PlatonicOctahedronItem::PlatonicOctahedronItem()
{
    m_edge.init("Edge", "Length of the edge", 16.0, Unit::nanometer, "edge");
}

std::unique_ptr<IFormFactor> PlatonicOctahedronItem::createFormFactor() const
{
    return std::make_unique<PlatonicOctahedron>(m_edge.value());
}

/* ------------------------------------------------ */

PlatonicTetrahedronItem::PlatonicTetrahedronItem()
{
    m_edge.init("Edge", "Length of the edge", 20.0, Unit::nanometer, "edge");
}

std::unique_ptr<IFormFactor> PlatonicTetrahedronItem::createFormFactor() const
{
    return std::make_unique<PlatonicTetrahedron>(m_edge.value());
}

/* ------------------------------------------------ */

DodecahedronItem::DodecahedronItem()
{
    m_edge.init("Edge", "Length of the edge", 7.0, Unit::nanometer, "edge");
}

std::unique_ptr<IFormFactor> DodecahedronItem::createFormFactor() const
{
    return std::make_unique<Dodecahedron>(m_edge.value());
}

/* ------------------------------------------------ */

IcosahedronItem::IcosahedronItem()
{
    m_edge.init("Edge", "Length of the edge", 10.0, Unit::nanometer, "edge");
}

std::unique_ptr<IFormFactor> IcosahedronItem::createFormFactor() const
{
    return std::make_unique<Icosahedron>(m_edge.value());
}
