//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/FormFactorItems.h
//! @brief     Defines FormFactorItems classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_SAMPLE_FORMFACTORITEMS_H
#define BORNAGAIN_GUI_MODEL_SAMPLE_FORMFACTORITEMS_H

#include "GUI/Model/Descriptor/DoubleProperty.h"
#include "GUI/Support/XML/Serialize.h"
#include <QVector>
#include <memory>

class IFormFactor;

#define FF_PROPERTY(nameLower, nameUpper)                                                          \
protected:                                                                                         \
    DoubleProperty m_##nameLower;                                                                  \
                                                                                                   \
public:                                                                                            \
    void set##nameUpper(double v)                                                                  \
    {                                                                                              \
        m_##nameLower.setValue(v);                                                                 \
    }

class FormFactorItem {
public:
    virtual ~FormFactorItem() = default;

public:
    virtual std::unique_ptr<IFormFactor> createFormFactor() const = 0;
    virtual DoublePropertyRefs geometryValues() = 0;
    virtual void serialize(Streamer& s) = 0;
};

class Pyramid2Item : public FormFactorItem {
public:
    Pyramid2Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(alpha, Alpha)

    DoublePropertyRefs geometryValues() override { return {m_length, m_width, m_height, m_alpha}; }
    void serialize(Streamer& s) override;
};

class BarGaussItem : public FormFactorItem {
public:
    BarGaussItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)

    DoublePropertyRefs geometryValues() override { return {m_length, m_width, m_height}; }
    void serialize(Streamer& s) override;
};

class BarLorentzItem : public FormFactorItem {
public:
    BarLorentzItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)

    DoublePropertyRefs geometryValues() override { return {m_length, m_width, m_height}; }
    void serialize(Streamer& s) override;
};

class BoxItem : public FormFactorItem {
public:
    BoxItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)

    DoublePropertyRefs geometryValues() override { return {m_length, m_width, m_height}; }
    void serialize(Streamer& s) override;
};

class ConeItem : public FormFactorItem {
public:
    ConeItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(alpha, Alpha)

    DoublePropertyRefs geometryValues() override { return {m_radius, m_height, m_alpha}; }
    void serialize(Streamer& s) override;
};

class Pyramid6Item : public FormFactorItem {
public:
    Pyramid6Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(baseEdge, BaseEdge)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(alpha, Alpha)

    DoublePropertyRefs geometryValues() override { return {m_baseEdge, m_height, m_alpha}; }
    void serialize(Streamer& s) override;
};

class Bipyramid4Item : public FormFactorItem {
public:
    Bipyramid4Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(heightRatio, HeightRatio)
    FF_PROPERTY(alpha, Alpha)

    DoublePropertyRefs geometryValues() override
    {
        return {m_length, m_height, m_heightRatio, m_alpha};
    }
    void serialize(Streamer& s) override;
};

class CylinderItem : public FormFactorItem {
public:
    CylinderItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)
    FF_PROPERTY(height, Height)

    DoublePropertyRefs geometryValues() override { return {m_radius, m_height}; }
    void serialize(Streamer& s) override;
};

class EllipsoidalCylinderItem : public FormFactorItem {
public:
    EllipsoidalCylinderItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radiusX, RadiusX)
    FF_PROPERTY(radiusY, RadiusY)
    FF_PROPERTY(height, Height)

    DoublePropertyRefs geometryValues() override { return {m_radiusX, m_radiusY, m_height}; }
    void serialize(Streamer& s) override;
};

class SphereItem : public FormFactorItem {
public:
    SphereItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)

    DoublePropertyRefs geometryValues() override { return {m_radius}; }
    void serialize(Streamer& s) override;
};

class SpheroidItem : public FormFactorItem {
public:
    SpheroidItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)
    FF_PROPERTY(height, Height)

    DoublePropertyRefs geometryValues() override { return {m_radius, m_height}; }
    void serialize(Streamer& s) override;
};

class HemiEllipsoidItem : public FormFactorItem {
public:
    HemiEllipsoidItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radiusX, RadiusX)
    FF_PROPERTY(radiusY, RadiusY)
    FF_PROPERTY(height, Height)

    DoublePropertyRefs geometryValues() override { return {m_radiusX, m_radiusY, m_height}; }
    void serialize(Streamer& s) override;
};

class Prism3Item : public FormFactorItem {
public:
    Prism3Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(baseEdge, BaseEdge)
    FF_PROPERTY(height, Height)

    DoublePropertyRefs geometryValues() override { return {m_baseEdge, m_height}; }
    void serialize(Streamer& s) override;
};

class Prism6Item : public FormFactorItem {
public:
    Prism6Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(baseEdge, BaseEdge)
    FF_PROPERTY(height, Height)

    DoublePropertyRefs geometryValues() override { return {m_baseEdge, m_height}; }
    void serialize(Streamer& s) override;
};

class Pyramid4Item : public FormFactorItem {
public:
    Pyramid4Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(baseEdge, BaseEdge)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(alpha, Alpha)

    DoublePropertyRefs geometryValues() override { return {m_baseEdge, m_height, m_alpha}; }
    void serialize(Streamer& s) override;
};

class CosineRippleBoxItem : public FormFactorItem {
public:
    CosineRippleBoxItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)

    DoublePropertyRefs geometryValues() override { return {m_length, m_width, m_height}; }
    void serialize(Streamer& s) override;
};

class CosineRippleGaussItem : public FormFactorItem {
public:
    CosineRippleGaussItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)

    DoublePropertyRefs geometryValues() override { return {m_length, m_width, m_height}; }
    void serialize(Streamer& s) override;
};

class CosineRippleLorentzItem : public FormFactorItem {
public:
    CosineRippleLorentzItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)

    DoublePropertyRefs geometryValues() override { return {m_length, m_width, m_height}; }
    void serialize(Streamer& s) override;
};

class SawtoothRippleBoxItem : public FormFactorItem {
public:
    SawtoothRippleBoxItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(asymmetry, Asymmetry)

    DoublePropertyRefs geometryValues() override
    {
        return {m_length, m_width, m_height, m_asymmetry};
    }
    void serialize(Streamer& s) override;
};

class SawtoothRippleGaussItem : public FormFactorItem {
public:
    SawtoothRippleGaussItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(asymmetry, Asymmetry)

    DoublePropertyRefs geometryValues() override
    {
        return {m_length, m_width, m_height, m_asymmetry};
    }
    void serialize(Streamer& s) override;
};

class SawtoothRippleLorentzItem : public FormFactorItem {
public:
    SawtoothRippleLorentzItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(asymmetry, Asymmetry)

    DoublePropertyRefs geometryValues() override
    {
        return {m_length, m_width, m_height, m_asymmetry};
    }
    void serialize(Streamer& s) override;
};

class Pyramid3Item : public FormFactorItem {
public:
    Pyramid3Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(baseEdge, BaseEdge)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(alpha, Alpha)

    DoublePropertyRefs geometryValues() override { return {m_baseEdge, m_height, m_alpha}; }
    void serialize(Streamer& s) override;
};

class TruncatedCubeItem : public FormFactorItem {
public:
    TruncatedCubeItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(removedLength, RemovedLength)

    DoublePropertyRefs geometryValues() override { return {m_length, m_removedLength}; }
    void serialize(Streamer& s) override;
};

class TruncatedSphereItem : public FormFactorItem {
public:
    TruncatedSphereItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(removedTop, RemovedTop)

    DoublePropertyRefs geometryValues() override { return {m_radius, m_height, m_removedTop}; }
    void serialize(Streamer& s) override;
};

class TruncatedSpheroidItem : public FormFactorItem {
public:
    TruncatedSpheroidItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(heightFlattening, HeightFlattening)
    FF_PROPERTY(removedTop, RemovedTop)

    DoublePropertyRefs geometryValues() override
    {
        return {m_radius, m_height, m_heightFlattening, m_removedTop};
    }
    void serialize(Streamer& s) override;
};

class CantellatedCubeItem : public FormFactorItem {
public:
    CantellatedCubeItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(removedLength, RemovedLength)

    DoublePropertyRefs geometryValues() override { return {m_length, m_removedLength}; }
    void serialize(Streamer& s) override;
};

class HorizontalCylinderItem : public FormFactorItem {
public:
    HorizontalCylinderItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)
    FF_PROPERTY(length, Length)
    FF_PROPERTY(sliceBottom, SliceBottom)
    FF_PROPERTY(sliceTop, SliceTop)

    DoublePropertyRefs geometryValues() override
    {
        return {m_radius, m_length, m_sliceBottom, m_sliceTop};
    }
    void serialize(Streamer& s) override;
};

// Platonic solids

class PlatonicItem : public FormFactorItem {
public:
    FF_PROPERTY(edge, Edge)

    DoublePropertyRefs geometryValues() override { return {m_edge}; }
    void serialize(Streamer& s) override;
};

class PlatonicOctahedronItem : public PlatonicItem {
public:
    PlatonicOctahedronItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;
};

class PlatonicTetrahedronItem : public PlatonicItem {
public:
    PlatonicTetrahedronItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;
};

class DodecahedronItem : public PlatonicItem {
public:
    DodecahedronItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;
};

class IcosahedronItem : public PlatonicItem {
public:
    IcosahedronItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;
};

#endif // BORNAGAIN_GUI_MODEL_SAMPLE_FORMFACTORITEMS_H
