//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/InterferenceItems.cpp
//! @brief     Implements InterferenceItems's classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/InterferenceItems.h"
#include "Base/Const/Units.h"
#include "GUI/Model/CatSample/Lattice2DItemCatalog.h"
#include "GUI/Model/CatSample/ProfileItemCatalogs.h"
#include "GUI/Model/Sample/Lattice2DItems.h"
#include "GUI/Model/Sample/ProfileItems.h"
#include "GUI/Support/XML/Serialize.h"
#include "Sample/Aggregate/Interferences.h"

namespace {
namespace Tag {

const QString PositionVariance("PositionVariance");
const QString Length("Length");
const QString RotationAngle("RotationAngle");
const QString IntegrateOverXi("IntegrateOverXi");
const QString DampingLength("DampingLength");
const QString DomainSize("DomainSize");
const QString DomainSize1("DomainSize1");
const QString DomainSize2("DomainSize2");
const QString Radius("Radius");
const QString Density("Density");
const QString PeakDistance("PeakDistance");
const QString Kappa("Kappa");

} // namespace Tag
} // namespace

InterferenceItem::InterferenceItem()
{
    m_positionVariance.init("PositionVariance", "Variance of the position in each dimension", 0.0,
                            Unit::nanometer2, "PositionVariance");
}

void InterferenceItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::PositionVariance, m_positionVariance);
}

// --------------------------------------------------------------------------------------------- //

Interference1DLatticeItem::Interference1DLatticeItem()
{
    m_length.init("Length", "Lattice length", 20.0, Unit::nanometer, "Length");
    m_rotationAngle.init(
        "Xi", "Rotation of lattice with respect to x-axis of reference frame (beam direction)", 0.0,
        Unit::degree, "xi");
    m_decayFunction.init<Profile1DItemCatalog>(
        "Decay Function", "One-dimensional decay function (finite size effects)", "decay");
}

std::unique_ptr<IInterference> Interference1DLatticeItem::createInterference() const
{
    auto result =
        std::make_unique<Interference1DLattice>(m_length, Units::deg2rad(m_rotationAngle));
    result->setDecayFunction(*m_decayFunction->createProfile());
    result->setPositionVariance(m_positionVariance);
    return std::unique_ptr<IInterference>(result.release());
}

void Interference1DLatticeItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::PositionVariance, m_positionVariance);
    Serialize::rwProperty(s, Tag::Length, m_length);
    Serialize::rwProperty(s, Tag::RotationAngle, m_rotationAngle);
    Serialize::rwSelected<Profile1DItemCatalog>(s, m_decayFunction);
}

// --------------------------------------------------------------------------------------------- //

Interference2DAbstractLatticeItem::Interference2DAbstractLatticeItem(bool xiIntegration)
    : m_xiIntegration(xiIntegration)
{
    m_latticeType.init<Lattice2DItemCatalog>("Lattice type", "", "latticeType");
    m_latticeType.set(new HexagonalLattice2DItem());
}

// --------------------------------------------------------------------------------------------- //

Interference2DLatticeItem::Interference2DLatticeItem()
    : Interference2DAbstractLatticeItem(false)
{
    m_decayFunction.init<Profile2DItemCatalog>(
        "Decay Function", "Two-dimensional decay function (finite size effects)", "decay");
}

std::unique_ptr<IInterference> Interference2DLatticeItem::createInterference() const
{
    Lattice2DItem* latticeItem = latticeType().currentItem();
    std::unique_ptr<Interference2DLattice> result(
        new Interference2DLattice(*latticeItem->createLattice()));

    result->setDecayFunction(*m_decayFunction->createProfile());
    result->setIntegrationOverXi(xiIntegration());
    result->setPositionVariance(m_positionVariance);

    return std::unique_ptr<IInterference>(result.release());
}

void Interference2DLatticeItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::PositionVariance, m_positionVariance);
    Serialize::rwValue(s, Tag::IntegrateOverXi, m_xiIntegration);
    Serialize::rwSelected<Lattice2DItemCatalog>(s, m_latticeType);
    Serialize::rwSelected<Profile2DItemCatalog>(s, m_decayFunction);
}

// --------------------------------------------------------------------------------------------- //

Interference2DParacrystalItem::Interference2DParacrystalItem()
    : Interference2DAbstractLatticeItem(true)
{
    m_dampingLength.init("Damping length", "The damping (coherence) length of the paracrystal", 0.0,
                         Unit::nanometer, "dampingLen");
    m_domainSize1.init("Domain size 1", "Size of the coherent domain along the first basis vector",
                       20000.0, Unit::nanometer, "size1");
    m_domainSize2.init("Domain size 2", "Size of the coherent domain along the second basis vector",
                       20000.0, Unit::nanometer, "size2");
    m_pdf1.init<Profile2DItemCatalog>(
        "PDF 1", "Probability distribution in first lattice direction", "pdf1");
    m_pdf2.init<Profile2DItemCatalog>(
        "PDF 2", "Probability distribution in second lattice direction", "pdf2");
}

std::unique_ptr<IInterference> Interference2DParacrystalItem::createInterference() const
{
    Lattice2DItem* latticeItem = latticeType().currentItem();

    std::unique_ptr<Interference2DParacrystal> result(
        new Interference2DParacrystal(*latticeItem->createLattice(), 0, 0, 0));

    result->setDampingLength(m_dampingLength);
    result->setDomainSizes(m_domainSize1, m_domainSize2);
    result->setIntegrationOverXi(xiIntegration());
    result->setProbabilityDistributions(*m_pdf1->createProfile(), *m_pdf2->createProfile());
    result->setPositionVariance(m_positionVariance);
    return std::unique_ptr<IInterference>(result.release());
}

void Interference2DParacrystalItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::PositionVariance, m_positionVariance);
    Serialize::rwValue(s, Tag::IntegrateOverXi, m_xiIntegration);
    Serialize::rwSelected<Lattice2DItemCatalog>(s, m_latticeType);
    Serialize::rwProperty(s, Tag::DampingLength, m_dampingLength);
    Serialize::rwProperty(s, Tag::DomainSize1, m_domainSize1);
    Serialize::rwProperty(s, Tag::DomainSize2, m_domainSize2);
    Serialize::rwSelected<Profile2DItemCatalog>(s, m_pdf1);
    Serialize::rwSelected<Profile2DItemCatalog>(s, m_pdf2);
}

// --------------------------------------------------------------------------------------------- //

InterferenceFinite2DLatticeItem::InterferenceFinite2DLatticeItem()
    : Interference2DAbstractLatticeItem(false)
{
}

std::unique_ptr<IInterference> InterferenceFinite2DLatticeItem::createInterference() const
{
    Lattice2DItem* latticeItem = latticeType().currentItem();
    auto result = std::make_unique<InterferenceFinite2DLattice>(*latticeItem->createLattice(),
                                                                m_domainSize1, m_domainSize2);

    result->setIntegrationOverXi(xiIntegration());
    result->setPositionVariance(m_positionVariance);

    return result;
}

void InterferenceFinite2DLatticeItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::PositionVariance, m_positionVariance);
    Serialize::rwValue(s, Tag::IntegrateOverXi, m_xiIntegration);
    Serialize::rwSelected<Lattice2DItemCatalog>(s, m_latticeType);
    Serialize::rwValue(s, Tag::DomainSize1, m_domainSize1);
    Serialize::rwValue(s, Tag::DomainSize1, m_domainSize2);
}

// --------------------------------------------------------------------------------------------- //

InterferenceHardDiskItem::InterferenceHardDiskItem()
{
    m_radius.init("Radius", "Hard disk radius", 5.0, Unit::nanometer, "radius");
    m_density.init("Total particle density", "Particle density in particles per area", 0.002,
                   Unit::nanometerMinus2, 5 /* decimals */, 0.0001 /* step */,
                   RealLimits::nonnegative(), "density");
}

std::unique_ptr<IInterference> InterferenceHardDiskItem::createInterference() const
{
    auto result = std::make_unique<InterferenceHardDisk>(m_radius, m_density);
    result->setPositionVariance(m_positionVariance);
    return std::unique_ptr<IInterference>(result.release());
}

void InterferenceHardDiskItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::PositionVariance, m_positionVariance);
    Serialize::rwProperty(s, Tag::Radius, m_radius);
    Serialize::rwProperty(s, Tag::Density, m_density);
}

// --------------------------------------------------------------------------------------------- //

InterferenceRadialParacrystalItem::InterferenceRadialParacrystalItem()
{
    m_peakDistance.init("Peak distance", "Average distance to the next neighbor", 20.0,
                        Unit::nanometer, "peak");
    m_dampingLength.init("Damping length", "The damping (coherence) length of the paracrystal",
                         1000.0, Unit::nanometer, "dampingLen");
    m_domainSize.init("Domain size", "Size of coherence domain along the lattice main axis", 0.0,
                      Unit::nanometer, "size");
    m_kappa.init("SizeSpaceCoupling",
                 "Size spacing coupling parameter of the Size Spacing Correlation Approximation",
                 0.0, Unit::unitless, "kappa");
    m_pdf.init<Profile1DItemCatalog>("PDF", "One-dimensional probability distribution", "pdf");
}

std::unique_ptr<IInterference> InterferenceRadialParacrystalItem::createInterference() const
{
    auto result = std::make_unique<InterferenceRadialParacrystal>(m_peakDistance, m_dampingLength);
    result->setDomainSize(m_domainSize);
    result->setKappa(m_kappa);
    auto pdf = m_pdf->createProfile();
    result->setProbabilityDistribution(*pdf);
    result->setPositionVariance(m_positionVariance);
    return std::unique_ptr<IInterference>(result.release());
}

void InterferenceRadialParacrystalItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::PositionVariance, m_positionVariance);
    Serialize::rwProperty(s, Tag::PeakDistance, m_peakDistance);
    Serialize::rwProperty(s, Tag::DampingLength, m_dampingLength);
    Serialize::rwProperty(s, Tag::DomainSize, m_domainSize);
    Serialize::rwProperty(s, Tag::Kappa, m_kappa);
    Serialize::rwSelected<Profile1DItemCatalog>(s, m_pdf);
}
