//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/MesocrystalItem.cpp
//! @brief     Implements class MesocrystalItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/MesocrystalItem.h"
#include "GUI/Model/CatSample/FormFactorItemCatalog.h"
#include "GUI/Model/CatSample/ItemWithParticlesCatalog.h"
#include "GUI/Model/CatSample/RotationItemCatalog.h"
#include "GUI/Model/Sample/CompoundItem.h"
#include "GUI/Model/Sample/CoreAndShellItem.h"
#include "GUI/Model/Sample/FormFactorItems.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Support/XML/Serialize.h"
#include "GUI/Util/Error.h"
#include "Sample/Particle/CoreAndShell.h"
#include "Sample/Particle/Crystal.h"
#include "Sample/Particle/IFormFactor.h"
#include "Sample/Particle/Mesocrystal.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Scattering/Rotations.h"

namespace {

namespace Tag {

const QString Abundance("Abundance");
const QString Position("Position");
const QString VectorA("VectorA");
const QString VectorB("VectorB");
const QString VectorC("VectorC");

} // namespace Tag

const QString abundance_tooltip = "Proportion of this type of mesocrystal normalized to the \n"
                                  "total number of particles in the layout";

const QString position_tooltip = "Relative position of the mesocrystal's reference point \n"
                                 "in the coordinate system of the parent (nm)";

} // namespace

MesocrystalItem::MesocrystalItem(const MaterialItems* materials)
    : ItemWithParticles(abundance_tooltip, position_tooltip)
    , m_materials(materials)
{
    m_vectorA.init("First lattice vector", "Coordinates of the first lattice vector",
                   Unit::nanometer, "vectorA");
    m_vectorB.init("Second lattice vector", "Coordinates of the second lattice vector",
                   Unit::nanometer, "vectorB");
    m_vectorC.init("Third lattice vector", "Coordinates of the third lattice vector",
                   Unit::nanometer, "vectorC");

    m_outerShape.init<FormFactorItemCatalog>("Outer Shape", "", "outerShape");
    m_basisParticle.init<ItemWithParticlesCatalog>("Basis", "", "basis", materials);
}

void MesocrystalItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Abundance, m_abundance);
    m_position.rwProperty(s, Tag::Position);
    Serialize::rwSelected<RotationItemCatalog>(s, m_rotation);
    m_vectorA.rwProperty(s, Tag::VectorA);
    m_vectorB.rwProperty(s, Tag::VectorB);
    m_vectorC.rwProperty(s, Tag::VectorC);
    Serialize::rwSelected<FormFactorItemCatalog>(s, m_outerShape);
    Serialize::rwSelected<ItemWithParticlesCatalog>(s, m_basisParticle, m_materials);
}

std::unique_ptr<Mesocrystal> MesocrystalItem::createMesocrystal() const
{
    const Lattice3D& lattice = getLattice();
    if (!(lattice.unitCellVolume() > 0.0))
        throw Error("MesocrystalItem::createMesocrystal(): "
                    "Lattice volume not strictly positive");
    std::unique_ptr<IParticle> basis = getBasis();
    if (!basis)
        throw Error("MesocrystalItem::createMesocrystal(): "
                    "No basis particle defined");
    Crystal crystal(*basis, lattice);

    std::unique_ptr<IFormFactor> ff = getOuterShape();
    if (!ff)
        throw Error("MesocrystalItem::createMesocrystal(): "
                    "No outer shape defined");

    auto result = std::make_unique<Mesocrystal>(crystal, *ff);
    result->setParticlePosition(position());
    if (auto r = createRotation(); r && !r->isIdentity())
        result->setRotation(*r);

    return result;
}

Lattice3D MesocrystalItem::getLattice() const
{
    return Lattice3D(m_vectorA, m_vectorB, m_vectorC);
}

std::unique_ptr<IParticle> MesocrystalItem::getBasis() const
{
    if (auto* p = dynamic_cast<ParticleItem*>(m_basisParticle.get()))
        return p->createParticle();

    if (auto* p = dynamic_cast<CoreAndShellItem*>(m_basisParticle.get()))
        return p->createCoreAndShell();

    if (auto* p = dynamic_cast<CompoundItem*>(m_basisParticle.get()))
        return p->createCompound();

    if (auto* p = dynamic_cast<MesocrystalItem*>(m_basisParticle.get()))
        return p->createMesocrystal();

    return {};
}

std::unique_ptr<IFormFactor> MesocrystalItem::getOuterShape() const
{
    return m_outerShape->createFormFactor();
}

ItemWithParticles* MesocrystalItem::basisParticle() const
{
    return m_basisParticle.get();
}

void MesocrystalItem::setBasis(ItemWithParticles* basis)
{
    m_basisParticle.set(basis);
}


QVector<ItemWithParticles*> MesocrystalItem::containedItemsWithParticles() const
{
    QVector<ItemWithParticles*> result;
    if (basisParticle())
        result << basisParticle() << basisParticle()->containedItemsWithParticles();
    return result;
}
