//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/MultiLayerItem.cpp
//! @brief     Implements class MultiLayerItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Support/XML/Backup.h"
#include "GUI/Support/XML/Serialize.h"
#include <QXmlStreamWriter>

namespace {
namespace Tag {

const QString Name("Name");
const QString Description("Description");
const QString CrossCorrelationLength("CrossCorrelationLength");
const QString Materials("Materials");
const QString Layers("Layers");
const QString ExternalField("ExternalField");

} // namespace Tag
} // namespace

MultiLayerItem::MultiLayerItem()
{
    m_name = "Sample";
    m_crossCorrelationLength.init(
        "Cross-correlation length", "Cross correlation length of roughnesses between interfaces",
        0.0, Unit::nanometer, 5 /* decimals */, RealLimits::nonnegative(), "cross");
    m_externalField.init("External field", "External field (A/m)", "A/m", "extField");
}

void MultiLayerItem::initFrom(const MultiLayerItem* other)
{
    GUI::Util::copyContents(other, this);
}

QVector<ItemWithMaterial*> MultiLayerItem::itemsWithMaterial() const
{
    QVector<ItemWithMaterial*> result;
    for (auto* layer : m_layers)
        result.append(layer->itemsWithMaterial());
    return result;
}

void MultiLayerItem::addStandardMaterials()
{
    m_materials.addRefractiveMaterial("Default", 1e-3, 1e-5);
    m_materials.addRefractiveMaterial("Vacuum", 0.0, 0.0);
    m_materials.addRefractiveMaterial("Particle", 6e-4, 2e-8);
    m_materials.addRefractiveMaterial("Substrate", 6e-6, 2e-8);
}

QVector<LayerItem*> MultiLayerItem::layers() const
{
    return m_layers;
}

LayerItem* MultiLayerItem::addLayer(int index)
{
    if (index < 0)
        index = m_layers.size();

    auto* layer = new LayerItem(&m_materials);
    m_layers.insert(index, layer);
    layer->setIsTopLayer(m_layers.first() == layer);
    layer->setIsBottomLayer(m_layers.last() == layer);
    return layer;
}

void MultiLayerItem::removeLayer(LayerItem* layer)
{
    m_layers.removeAll(layer);
    delete layer;

    if (!m_layers.isEmpty()) {
        m_layers.first()->setIsTopLayer(true);
        m_layers.last()->setIsBottomLayer(true);
    }
}

void MultiLayerItem::moveLayer(LayerItem* layer, LayerItem* beforeThisLayer)
{
    m_layers.removeAll(layer);
    int index = m_layers.size(); // move to end
    if (beforeThisLayer != nullptr)
        index = m_layers.indexOf(beforeThisLayer);
    m_layers.insert(index, layer);
    for (auto* l : m_layers) {
        l->setIsTopLayer(m_layers.first() == l);
        l->setIsBottomLayer(m_layers.last() == l);
    }
}

void MultiLayerItem::writeTo(QXmlStreamWriter* writer) const
{
    Streamer s(writer);
    const_cast<MultiLayerItem*>(this)->serialize(s);
}

void MultiLayerItem::readFrom(QXmlStreamReader* reader)
{
    Streamer s(reader);
    serialize(s);
}

void MultiLayerItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwValue(s, Tag::Name, m_name);
    Serialize::rwValue(s, Tag::Description, m_description);
    Serialize::rwProperty(s, Tag::CrossCorrelationLength, m_crossCorrelationLength);
    m_externalField.rwProperty(s, Tag::ExternalField);
    Serialize::rwClass(s, Tag::Materials, m_materials);
    Serialize::rwVector(s, Tag::Layers, m_layers, &m_materials);

    if (s.xmlReader()) {
        // set non-stored infos
        for (auto* l : m_layers) {
            l->setIsTopLayer(m_layers.first() == l);
            l->setIsBottomLayer(m_layers.last() == l);
        }
    }
}
