//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/Processed/Slicer.cpp
//! @brief     Implements function Compute::Slicing::sliceFormFactor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Resample/Processed/Slicer.h"
#include "Base/Math/Constants.h"
#include "Base/Math/Functions.h"
#include "Base/Util/Assert.h"
#include "Base/Vector/RotMatrix.h"
#include "Resample/Particle/ReCompound.h"
#include "Resample/Particle/ReMesocrystal.h"
#include "Resample/Particle/ReParticle.h"
#include "Resample/Slice/ZLimits.h"
#include "Sample/HardParticle/HardParticles.h"
#include "Sample/Material/MaterialFactoryFuncs.h"
#include "Sample/Material/MaterialUtils.h"
#include "Sample/Particle/Compound.h"
#include "Sample/Particle/CoreAndShell.h"
#include "Sample/Particle/Crystal.h"
#include "Sample/Particle/Mesocrystal.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Scattering/Rotations.h"
#include <iostream>

namespace {

//! Nested structure that holds slicing effects on position and removed parts.
struct SlicingEffects {
    R3 position;
    double dz_bottom;
    double dz_top;
};

SlicingEffects computeSlicingEffects(ZLimits limits, const R3& position, double height)
{
    R3 new_position(position);
    double z_bottom = position.z();
    double z_top = position.z() + height;
    double dz_top = std::isinf(limits.zTop()) ? -1 : z_top - limits.zTop();
    double dz_bottom = std::isinf(limits.zBottom()) ? -1 : limits.zBottom() - z_bottom;
    ASSERT(dz_top >= 0 || dz_bottom >= 0);
    ASSERT(dz_bottom <= height);
    ASSERT(dz_top <= height);
    if (dz_bottom < 0)
        dz_bottom = 0;
    if (dz_top < 0)
        dz_top = 0;
    if (dz_bottom > 0)
        new_position.setZ(limits.zBottom());
    return {new_position, dz_bottom, dz_top};
}

bool shapeIsContainedInLimits(const IFormFactor& formfactor, ZLimits limits,
                              const IRotation* rotation, R3 translation)
{
    double zbottom = formfactor.bottomZ(rotation) + translation.z();
    double ztop = formfactor.topZ(rotation) + translation.z();

    return limits.zBottom() <= zbottom && ztop <= limits.zTop();
}

bool shapeOutsideLimits(const IFormFactor& formfactor, ZLimits limits, const IRotation* rotation,
                        R3 translation)
{
    double zbottom = formfactor.bottomZ(rotation) + translation.z();
    double ztop = formfactor.topZ(rotation) + translation.z();

    return ztop <= limits.zBottom() || zbottom >= limits.zTop();
}

ReParticle* createTransformedFormFactor(const IFormFactor* formfactor, const IRotation* rot,
                                        R3 translation)
{
    auto* result = new ReParticle(*formfactor);
    if (rot && !rot->isIdentity())
        result->setRotMatrix(rot->rotMatrix());
    if (translation != R3())
        result->setPosition(translation);
    return result;
}

ReParticle* createParticleSlice(const IFormFactor* ff, ZLimits limits, const IRotation* rot,
                                R3 translation)
{
    if (shapeOutsideLimits(*ff, limits, rot, translation))
        return nullptr;

    if (shapeIsContainedInLimits(*ff, limits, rot, translation))
        return createTransformedFormFactor(ff, rot, translation);

    if (!ff->canSliceAnalytically(rot))
        throw std::runtime_error("Slicing of " + ff->className()
                                 + " not supported for the given rotation");

    if (const auto* f = dynamic_cast<const Pyramid2*>(ff)) {
        const SlicingEffects effects = computeSlicingEffects(limits, translation, f->height());
        double dbase_edge = 2 * effects.dz_bottom * Math::cot(f->alpha());
        auto slicedff = new Pyramid2(f->length() - dbase_edge, f->width() - dbase_edge,
                                     f->height() - effects.dz_bottom - effects.dz_top, f->alpha());
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Box*>(ff)) {
        const SlicingEffects effects = computeSlicingEffects(limits, translation, f->height());
        auto slicedff =
            new Box(f->length(), f->width(), f->height() - effects.dz_bottom - effects.dz_top);
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Cone*>(ff)) {
        const SlicingEffects effects = computeSlicingEffects(limits, translation, f->height());
        double dradius = effects.dz_bottom * Math::cot(f->alpha());
        auto slicedff = new Cone(f->radius() - dradius,
                                 f->height() - effects.dz_bottom - effects.dz_top, f->alpha());
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Pyramid6*>(ff)) {
        const SlicingEffects effects = computeSlicingEffects(limits, translation, f->height());
        double dbase_edge = effects.dz_bottom * Math::cot(f->alpha());
        auto slicedff = new Pyramid6(f->baseEdge() - dbase_edge,
                                     f->height() - effects.dz_bottom - effects.dz_top, f->alpha());
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Bipyramid4*>(ff)) {
        const SlicingEffects effects =
            computeSlicingEffects(limits, translation, f->height() * (1 + f->heightRatio()));
        if (effects.dz_bottom > f->height()) {
            double dbase_edge = 2 * (effects.dz_bottom - f->height()) * Math::cot(f->alpha());
            auto slicedff = new Pyramid4(f->length() - dbase_edge,
                                         f->height() * (1 + f->heightRatio()) - effects.dz_bottom
                                             - effects.dz_top,
                                         f->alpha());
            return createTransformedFormFactor(slicedff, rot, effects.position);
        }
        if (effects.dz_top > f->heightRatio() * f->height()) {
            double dbase_edge = 2 * (f->height() - effects.dz_bottom) * Math::cot(f->alpha());
            auto slicedff = new Pyramid4(f->length() - dbase_edge,
                                         f->height() * (1 + f->heightRatio()) - effects.dz_bottom
                                             - effects.dz_top,
                                         M_PI - f->alpha());
            return createTransformedFormFactor(slicedff, rot, effects.position);
        }
        auto slicedff = new Bipyramid4(f->length(), f->height() - effects.dz_bottom,
                                       f->heightRatio() * f->height() - effects.dz_top, f->alpha());
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Cylinder*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        auto slicedff = new Cylinder(f->radius(), f->height() - effects.dz_bottom - effects.dz_top);
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const EllipsoidalCylinder*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        auto slicedff = new EllipsoidalCylinder(f->radiusX(), f->radiusY(),
                                                f->height() - effects.dz_bottom - effects.dz_top);
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const HorizontalCylinder*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, 2 * f->radius());
        auto slicedff =
            new HorizontalCylinder(f->radius(), f->length(), -f->radius() + effects.dz_bottom,
                                   f->radius() - effects.dz_top);
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Sphere*>(ff)) {
        R3 new_translation = translation;
        if (rot) {
            R3 center(0.0, 0.0, f->radius());
            new_translation += rot->transformed(center) - center;
        }
        const double h = 2 * f->radius();
        auto effects = computeSlicingEffects(limits, translation, h);
        auto slicedff = new TruncatedSphere(f->radius(), h - effects.dz_bottom, effects.dz_top);
        return createTransformedFormFactor(slicedff, nullptr, effects.position);

    } else if (const auto* f = dynamic_cast<const Spheroid*>(ff)) {
        double flattening = f->height() / (2.0 * f->radius());
        auto effects = computeSlicingEffects(limits, translation, f->height());
        auto slicedff = new TruncatedSpheroid(f->radius(), f->height() - effects.dz_bottom,
                                              flattening, effects.dz_top);
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const LongBoxGauss*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        auto slicedff = new LongBoxGauss(f->length(), f->width(),
                                         f->height() - effects.dz_bottom - effects.dz_top);
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const LongBoxLorentz*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        auto slicedff = new LongBoxLorentz(f->length(), f->width(),
                                           f->height() - effects.dz_bottom - effects.dz_top);
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Prism3*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        auto slicedff = new Prism3(f->baseEdge(), f->height() - effects.dz_bottom - effects.dz_top);
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Prism6*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        auto slicedff = new Prism6(f->baseEdge(), f->height() - effects.dz_bottom - effects.dz_top);
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Pyramid4*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        double dbaseEdge = 2 * effects.dz_bottom * Math::cot(f->alpha());
        auto slicedff = new Pyramid4(f->baseEdge() - dbaseEdge,
                                     f->height() - effects.dz_bottom - effects.dz_top, f->alpha());
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Pyramid3*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        double dbaseEdge = 2 * sqrt(3) * effects.dz_bottom * Math::cot(f->alpha());
        auto slicedff = new Pyramid3(f->baseEdge() - dbaseEdge,
                                     f->height() - effects.dz_bottom - effects.dz_top, f->alpha());
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const TruncatedSphere*>(ff)) {
        double height = f->height() - f->removedTop();
        auto effects = computeSlicingEffects(limits, translation, height);
        auto slicedff = new TruncatedSphere(f->radius(), f->height() - effects.dz_bottom,
                                            effects.dz_top + f->removedTop());
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const TruncatedSpheroid*>(ff)) {
        double height = f->height() - f->removedTop();
        auto effects = computeSlicingEffects(limits, translation, height);
        auto slicedff =
            new TruncatedSpheroid(f->radius(), f->height() - effects.dz_bottom,
                                  f->heightFlattening(), effects.dz_top + f->removedTop());
        return createTransformedFormFactor(slicedff, rot, effects.position);

    } else
        throw std::runtime_error("Slicing of " + ff->className() + " not supported");
}

//! Recursively processes the basis of a mesocrystal.
IReParticle* processBasis(const IParticle* basis, const Material& ambientMat)
{
    if (const auto* b = dynamic_cast<const Compound*>(basis)) {
        const auto& particles = b->decompose();
        ASSERT(!particles.empty());
        auto* result = new ReCompound;
        for (const auto* particle : particles) {
            std::unique_ptr<IReParticle> re(processBasis(particle, ambientMat));
            result->addFormFactor(*re);
        }
        return result;
    } else if (dynamic_cast<const CoreAndShell*>(basis)) {
        throw std::runtime_error("Mesocrystal with CoreAndShell basis not yet supported");
    }
    // Remaining case: the basis is a single particle.
    return Compute::Slicing::particlesInSlice(basis, {}, ambientMat).release_back();
}

} // namespace


OwningVector<IReParticle> Compute::Slicing::particlesInSlice(const IParticle* particle,
                                                             const ZLimits& limits,
                                                             const Material& ambientMat)
{
    if (const auto* p = dynamic_cast<const Particle*>(particle)) {
        ASSERT(p->formfactor_at_bottom());
        std::unique_ptr<ReParticle> particleSlice(createParticleSlice(
            p->formfactor_at_bottom(), limits, p->rotation(), p->particlePosition()));
        if (!particleSlice)
            return {};
        double volume = particleSlice->volume();
        Material transformed_material(
            p->rotation() ? p->material()->rotatedMaterial(p->rotation()->rotMatrix())
                          : *p->material());
        particleSlice->setMaterial(transformed_material);
        particleSlice->setAdmixedFraction(volume);
        particleSlice->setAdmixedMaterial(transformed_material);
        particleSlice->setAmbientMaterial(ambientMat);
        OwningVector<IReParticle> result;
        result.emplace_back(particleSlice.release());
        return result;

    } else if (const auto* p = dynamic_cast<const CoreAndShell*>(particle)) {
        const Particle* core = p->coreParticle();
        const Particle* shell = p->shellParticle();
        ASSERT(core && shell);

        // shell
        std::unique_ptr<Particle> P_shell(shell->clone());
        if (p->rotation())
            P_shell->rotate(*p->rotation());
        P_shell->translate(p->particlePosition());
        OwningVector<IReParticle> shell_slices =
            particlesInSlice(P_shell.get(), limits, ambientMat);
        if (shell_slices.empty())
            return {};
        ASSERT(shell_slices.size() == 1);
        IReParticle* sliced_shell = shell_slices.release_back();
        if (!sliced_shell)
            return {};
        const Material& shell_material = sliced_shell->admixed().material;

        // core
        std::unique_ptr<Particle> P_core(core->clone());
        if (p->rotation())
            P_core->rotate(*p->rotation());
        P_core->translate(p->particlePosition());
        OwningVector<IReParticle> core_slices =
            particlesInSlice(P_core.get(), limits, shell_material);
        ASSERT(core_slices.size() == 1);
        IReParticle* sliced_core = core_slices.release_back();

        // if core out of limits, return sliced shell
        if (!sliced_core) {
            OwningVector<IReParticle> result;
            result.emplace_back(sliced_shell);
            return result;
        }

        sliced_shell->setAdmixedFraction(sliced_shell->admixedFraction()
                                         - sliced_core->admixedFraction());

        OwningVector<IReParticle> result;
        result.emplace_back(sliced_shell);
        result.emplace_back(sliced_core);
        return result;

    } else if (dynamic_cast<const Compound*>(particle)) {
        throw std::runtime_error("Compound does not yet support slicing");

    } else if (const auto* p = dynamic_cast<const Mesocrystal*>(particle)) {
        const Crystal* crystal = &p->particleStructure();
        const IFormFactor* meso_formfactor = p->outerShape();
        ASSERT(crystal && meso_formfactor);

        double unit_cell_volume = crystal->lattice()->unitCellVolume();
        if (unit_cell_volume <= 0)
            return {};

        std::unique_ptr<ReParticle> new_shape(
            createParticleSlice(meso_formfactor, limits, p->rotation(), p->particlePosition()));

        const std::unique_ptr<Crystal> new_crystal(
            crystal->transformed(p->rotation(), p->particlePosition()));

        const std::unique_ptr<IReParticle> new_basis(
            processBasis(new_crystal->basis(), ambientMat));

        auto mc = std::make_unique<ReMesocrystal>(std::optional<size_t>{}, *new_crystal->lattice(),
                                                  *new_basis, *new_shape,
                                                  new_crystal->position_variance());

        std::vector<double> weights;
        std::vector<Material> materials;
        double volume = 0;
        for (const auto* p : crystal->basis()->decompose()) {
            volume += p->volume();
            weights.push_back(p->volume());
            materials.push_back(p->avgeMaterial());
        }
        mc->setAdmixedFraction(volume / unit_cell_volume);
        mc->setAdmixedMaterial(
            MaterialUtils::averagedMaterial("MesocrystalAvgeMat", weights, materials));

        OwningVector<IReParticle> result;
        result.emplace_back(mc.release());
        return result;

    } else
        ASSERT(0);
}

ZLimits Compute::Slicing::zSpan(const IParticle* particle)
{
    if (const auto* p = dynamic_cast<const Particle*>(particle)) {
        ASSERT(p->formfactor_at_bottom());
        std::unique_ptr<ReParticle> particleSlice(createParticleSlice(
            p->formfactor_at_bottom(), ZLimits(), p->rotation(), p->particlePosition()));
        if (!particleSlice)
            return {};
        return particleSlice->Z_span();

    } else if (const auto* p = dynamic_cast<const Compound*>(particle)) {
        const auto subparticles = p->decompose();
        ASSERT(subparticles.size() > 0);
        ZLimits result = zSpan(subparticles[0]);
        for (const auto* subparticle : subparticles)
            // TODO why no rotation and translation as for CoreAndShell ??
            result = ZLimits::enclosingInterval(result, zSpan(subparticle));
        return result;

    } else if (const auto* p = dynamic_cast<const CoreAndShell*>(particle)) {

        std::unique_ptr<Particle> pp(p->shellParticle()->clone());

        if (pp->rotation())
            pp->rotate(*p->rotation());
        pp->translate(p->particlePosition());

        return zSpan(pp.get());

    } else if (const auto* p = dynamic_cast<const Mesocrystal*>(particle)) {
        const IFormFactor* meso_formfactor = p->outerShape();
        ASSERT(meso_formfactor);
        std::unique_ptr<ReParticle> new_shape(
            createParticleSlice(meso_formfactor, ZLimits(), p->rotation(), p->particlePosition()));

        return {new_shape->bottom_Z(nullptr), new_shape->top_Z(nullptr)};

    } else
        ASSERT(0);
}
